/* ======================================================================== */
/*  CP-1600 Advanced Disassembler                                           */
/*  By Joseph Zbiciak                                                       */
/* ------------------------------------------------------------------------ */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ------------------------------------------------------------------------ */
/*                 Copyright (c) 1998-2001, Joseph Zbiciak                  */
/* ======================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "config.h"
#include "macros.h"
#include "icart/icartrom.h"
#include "icart/icartbin.h"

/* ======================================================================== */
/*  These are errors that can be reported by the Intellicart routines.      */
/* ======================================================================== */
static char *rom_errors[] =
{
    "No Error",
    "Bad Arguments",
    "Bad ROM Header",
    "CRC-16 Error in ROM Segments",
    "Bad ROM Segment Address Range",
    "Bad ROM Fine-Address Range",   
    "CRC-16 Error in Enable Tables",
    "Unknown Error"
};

icartrom_t the_icart;
symtab_t   *symtab;

#define MASK_DATA   (0x0003)    /* lower two bits indicate data type        */
#define FLAG_DATA   (0x0001)    /* data at this address is DECLE            */
#define FLAG_DBDATA (0x0002)    /* data at this and next addr is BIDECLE    */
#define FLAG_STRING (0x0003)    /* data at this address is STRING           */
#define FLAG_CODE   (0x0004)    /* this address is presumed to be CODE      */
#define FLAG_INVOP  (0x0008)    /* opcode at this addr is invalid.          */
#define FLAG_BRANCH (0x0010)    /* J, JSR, or B, or arith on PC             */
#define FLAG_CONDBR (0x0020)    /* This branch is conditional.              */
#define FLAG_JSR    (0x0040)    /* Further distinguish JSR from J, B        */
#define FLAG_SDBD   (0x0080)    /* Instruction is modified by SDBD          */
#define FLAG_BRTRG  (0x0100)    /* Instruction is known branch target.      */
#define FLAG_EMPTY  (0x80000000) /* nothing loaded at this address.         */

#define BTARG_NONE  (0)         /* no branch target for instruction         */
#define BTARG_MULT  (0x20000)   /* This instr. has multiple br targets.     */
#define BTARG_UNK   (0x40000)   /* branch target is unknown                 */

#define OPF_SRC     (0x0001)    /* This operand is a 'source' operand.      */
#define OPF_DST     (0x0002)    /* This operand is a 'destination' operand. */
#define OPF_SRCDST  (OPF_SRC|OPF_DST)
#define OPF_IND     (0x0004)    /* Value accessed thru operand indirectly   */
#define OPF_ADDR    (0x0008)    /* This operand represents an address.      */

/* ======================================================================== */
/*  INSTRUCTION MNEMONICS                                                   */
/* ======================================================================== */
typedef enum mnm_t
{
    M_err = 0,
    M_HLT,  M_SDBD, M_EIS,  M_DIS,          M_TCI,  M_CLRC, M_SETC,
    M_JSR,  M_JSRE, M_JSRD,         M_J,    M_JE,   M_JD,
    M_SWAP, M_SLL,  M_RLC,  M_SLLC, M_SLR,  M_SAR,  M_RRC,  M_SARC,
    M_B,    M_BC,   M_BOV,  M_BPL,  M_BEQ,  M_BLT,  M_BLE,  M_BUSC,
    M_NOPP, M_BNC,  M_BNOV, M_BMI,  M_BNEQ, M_BGE,  M_BGT,  M_BESC,
    M_BEXT,
            M_MVO,  M_MVI,  M_ADD,  M_SUB,  M_CMP,  M_AND,  M_XOR,
            M_MVOI, M_MVII, M_ADDI, M_SUBI, M_CMPI, M_ANDI, M_XORI,
            M_MVO_, M_MVI_, M_ADD_, M_SUB_, M_CMP_, M_AND_, M_XOR_,
                    M_MOVR, M_ADDR, M_SUBR, M_CMPR, M_ANDR, M_XORR,
            M_INCR, M_DECR, M_COMR, M_NEGR, M_ADCR,         M_RSWD,

    M_DECLE, M_BIDECLE
};

char *mnemonics[] =
{
    "err!",
    "HLT",  "SDBD", "EIS",  "DIS",          "TCI",  "CLRC", "SETC",
    "JSR",  "JSRE", "JSRD",         "J",    "JE",   "JD",
    "SWAP", "SLL",  "RLC",  "SLLC", "SLR",  "SAR",  "RRC",  "SARC",
    "B",    "BC",   "BOV",  "BPL",  "BEQ",  "BLT",  "BLE",  "BUSC",
    "NOPP", "BNC",  "BNOV", "BMI",  "BNEQ", "BGE",  "BGT",  "BESC",
    "BEXT",
            "MVO",  "MVI",  "ADD",  "SUB",  "CMP",  "AND",  "XOR",
            "MVOI", "MVII", "ADDI", "SUBI", "CMPI", "ANDI", "XORI",
            "MVO@", "MVI@", "ADD@", "SUB@", "CMP@", "AND@", "XOR@",
                    "MOVR", "ADDR", "SUBR", "CMPR", "ANDR", "XORR",
            "INCR", "DECR", "COMR", "NEGR", "ADCR",         "RSWD", 

    "DECLE", "BIDECLE"
};


mnm_t mnm_dir_2op[8] =
{
    M_err,  M_MVO,  M_MVI,  M_ADD,  M_SUB,  M_CMP,  M_AND,  M_XOR
};

mnm_t mnm_ind_2op[8] =
{
    M_err,  M_MVO_, M_MVI_, M_ADD_, M_SUB_, M_CMP_, M_AND_, M_XOR_
};

mnm_t mnm_imm_2op[8] =
{
    M_err,  M_MVOI, M_MVII, M_ADDI, M_SUBI, M_CMPI, M_ANDI, M_XORI
};

mnm_t mnm_rot_1op[8] = 
{
    M_SWAP, M_SLL,  M_RLC,  M_SLLC, M_SLR,  M_SAR,  M_RRC,  M_SARC
};

mnm_t opt_reg_1op[8] =
{
    M_err,  M_INCR, M_DECR, M_COMR, M_NEGR, M_ADCR, M_err,  M_RSWD
};

mnm_t mnm_cond_br[16] =
{
    M_B,    M_BC,   M_BOV,  M_BPL,  M_BEQ,  M_BLT,  M_BLE,  M_BUSC, 
    M_NOPP, M_BNC,  M_BNOV, M_BMI,  M_BNEQ, M_BGE,  M_BGT,  M_BESC
};

mnm_t mnm_impl_1op_a[4] = { M_HLT,  M_SDBD, M_EIS,  M_DIS  };
mnm_t mnm_impl_1op_b[4] = { M_err,  M_TCI,  M_CLRC, M_SETC };

mnm_t mnm_jsr[8] = 
{
    M_JSR,  M_JSRE, M_JSRD, M_err,  M_J,    M_JE,   M_JD,   M_err 
};

/* ======================================================================== */
/*  Note:  Some attempt is made to avoid namespace collision w/ jzIntv      */
/*  source code names, in case I re-integrate this w/ jzIntv someday.       */
/* ======================================================================== */
enum oper_type 
{ 
    OP_NONE = 0,                /* No operand.                              */
    OP_REG,                     /* CPU register.                            */
    OP_IMMED,                   /* Immediate constant.                      */
    OP_DIRADDR,                 /* Operand stored at indicated address.     */
    OP_BRTRG                    /* Branch target (special immediate).       */
};

typedef struct dis_oper_t
{
    oper_type   type;
    uint_16     op;             /* address, reg# or value                   */
    uint_16     flags;          /* flags associated w/ operand.             */
} dis_oper_t;

typedef struct dis_instr_t
{
    int      len;               /* length of this instruction               */
    mnm_t    mnemonic;          /* Mnemonic for this instruction.           */
    char    *label;             /* label applied to this instruction        */
    char    *cmt;               /* comment(s) associated w/ address.        */
    char    *fmtline;           /* formatted line for this instruction.     */
    oper_t   op1;               /* first operand.                           */
    oper_t   op2;               /* second operand.                          */
    unsigned flags;             /* Other information about this instruction */
    int      br_target;         /* branch target address, if known          */
    /* todo -- have a way of recording a list of branch targets if multiple */
} dis_instr_t;


/* ======================================================================== */
/*  INSTRUCTION PRINTER TABLE (forward references)                          */
/* ======================================================================== */
typedef char *(*ins_prt_t)(uint_32);

char *prt_err(uint_32);
char *prt_imp(uint_32);
char *prt_jsr(uint_32);
char *prt_cbr(uint_32);
char *prt_bxt(uint_32);
char *prt_dir(uint_32);
char *prt_imm(uint_32);
char *prt_dro(uint_32);
char *prt_imo(uint_32);
char *prt_2rg(uint_32);
char *prt_1rg(uint_32);
char *prt_dcl(uint_32);
char *prt_bid(uint_32);

ins_prt_t instr_printers[] =
{
/*  "err!"  */
    prt_err,

/*  "HLT",  "SDBD", "EIS",  "DIS",          "TCI",  "CLRC", "SETC", */
    prt_imp,prt_imp,prt_imp,prt_imp,        prt_imp,prt_imp,prt_imp,

/*  "JSR",  "JSRE", "JSRD",         "J",    "JE",   "JD",           */
    prt_jsr,prt_jsr,prt_jsr,        prt_cbr,prt_cbr,prt_cbr,

/*  "SWAP", "SLL",  "RLC",  "SLLC", "SLR",  "SAR",  "RRC",  "SARC", */
    prt_rot,prt_rot,prt_rot,prt_rot,prt_rot,prt_rot,prt_rot,prt_rot,

/*  "B",    "BC",   "BOV",  "BPL",  "BEQ",  "BLT",  "BLE",  "BUSC", */
    prt_cbr,prt_cbr,prt_cbr,prt_cbr,prt_cbr,prt_cbr,prt_cbr,prt_cbr,
/*  "NOPP", "BNC",  "BNOV", "BMI",  "BNEQ", "BGE",  "BGT",  "BESC", */
    prt_imp,prt_cbr,prt_cbr,prt_cbr,prt_cbr,prt_cbr,prt_cbr,prt_cbr,
/*  "BEXT",                                                         */
    prt_bxt,

/*          "MVO",  "MVI",  "ADD",  "SUB",  "CMP",  "AND",  "XOR",  */
            prt_jsr,prt_dir,prt_dir,prt_dir,prt_dir,prt_dir,prt_dir,
/*          "MVOI", "MVII", "ADDI", "SUBI", "CMPI", "ANDI", "XORI", */
            prt_imo,prt_imm,prt_imm,prt_imm,prt_imm,prt_imm,prt_imm,
/*          "MVO@", "MVI@", "ADD@", "SUB@", "CMP@", "AND@", "XOR@", */
            prt_2rg,prt_2rg,prt_2rg,prt_2rg,prt_2rg,prt_2rg,prt_2rg,
/*                  "MOVR", "ADDR", "SUBR", "CMPR", "ANDR", "XORR", */
                    prt_2rg,prt_2rg,prt_2rg,prt_2rg,prt_2rg,prt_2rg,
/*          "INCR", "DECR", "COMR", "NEGR", "ADCR",         "RSWD", */
            prt_1rg,prt_1rg,prt_1rg,prt_1rg,prt_1rg,        prt_1rg,

/*  "DECLE", "BIDECLE"                                              */
    prt_dcl, prt_bid
};


/* ======================================================================== */
/*  Declare storage for the ROM image.                                      */
/* ======================================================================== */
icartrom_t icart;
dis_instr_t instr[65536];  /* Yes, an instr for every address.  why not? */




/* ======================================================================== */
/*  HELPER FUNCTIONS AND MACROS                                             */
/* ======================================================================== */


#define INSTR_ADDR(i) ((i) - instr)
#define GET_WORD(a)   (icart.image[(a) & 0xFFFF])
#define NOT_CODE(a)   ((instr[(a)].flags & (FLAG_EMPTY|FLAG_INVOP|MASK_DATA)))
#define IS_EMPTY(a)   ((instr[(a)].flags & (FLAG_EMPTY)) != 0)
#define MAYBE_CODE(a) (!NOT_CODE(a))
#define BAD_ARG10(a)  ((GET_WORD(a) & 0xFC00) || IS_EMPTY(a))
#define IS_BRANCH(a)  ((instr[(a)].flags & FLAG_BRANCH) != 0)
#define IS_CONDBR(a)  ((instr[(a)].flags & FLAG_CONDBR) != 0)

#define GET_BIT(bv,i,b) do {                                    \
                            int _ = (i);                        \
                            b = ((bv)[_ >> 5] >> (_ & 31)) & 1; \
                        } while(0)


#define VB_PRINTF(v,x)  do {                                    \
                            if (verbose > (v))                  \
                            {                                   \
                                printf x;                       \
                                fflush(stdout);                 \
                            }                                   \
                        } while(0)


/* ======================================================================== */
/* ======================================================================== */
/*  ANALYSIS PASSES AND DECODER FUNCTIONS                                   */
/* ======================================================================== */
/* ======================================================================== */
int hlt_is_invalid                = 1;
int rare_ops_are_invalid          = 1;
int allow_branch_target_wrap      = 0;
int suspicious_pc_math_is_invalid = 1;
int verbose                       = 0;


/* ======================================================================== */
/*  DIS_JUMP -- decode J/JE/JD/JSR/JSRE/JSRD                                */
/* ======================================================================== */
void dis_jump      (uint_32 addr)
{
    uint_16 w0, w1, w2;
    uint_16 btarg;
    int     ret_reg;
    int     int_mode;

    /* -------------------------------------------------------------------- */
    /*  Make sure all three words are in implemented memory and are 10-bit. */
    /* -------------------------------------------------------------------- */
    if (BAD_ARG10(addr + 1) || BAD_ARG10(addr + 2))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }
        
    /* -------------------------------------------------------------------- */
    /*  Get all three words of the jump instruction.                        */
    /* -------------------------------------------------------------------- */
    w0 = GET_WORD(addr);
    w1 = GET_WORD(addr + 1);
    w2 = GET_WORD(addr + 2);

    /* -------------------------------------------------------------------- */
    /*  Decode "interrupt mode".  This is in bits 0 and 1 of word 2.        */
    /*      00 means "No Change."                                           */
    /*      01 means "Jump and Enable Interrupts"                           */
    /*      10 means "Jump and Disable Interrupts"                          */
    /*      11 is an invalid opcode.                                        */
    /* -------------------------------------------------------------------- */
    int_mode = (w2 & 3);
    if (w2 == 3)
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Decode return register.  This is in bits 8 and 9 of word 2.         */
    /*      00 means R4         10 means R6                                 */
    /*      01 means R5         11 means "Do not save return address."      */
    /* -------------------------------------------------------------------- */
    ret_reg = 4 + ((w2 & 0x300) >> 8);

    /* -------------------------------------------------------------------- */
    /*  Decode branch target.  Lower 10 bits come from third word.  Upper   */
    /*  6 bits come from bits 2 thru 7 of second word.                      */
    /* -------------------------------------------------------------------- */
    btarg = ((w2 & 0x0FC) << 8) | (w3 & 0x3FF);

    /* -------------------------------------------------------------------- */
    /*  Generate the decoded instruction.                                   */
    /* -------------------------------------------------------------------- */
    instr[addr].len       = 3;
    instr[addr].br_target = btarg;
    instr[addr].flags    |= FLAG_BRANCH;

    if (ret_reg < 7)  /* JSR/JSRE/JSRD */
    {
        instr[addr].op1.type  = OP_REG;
        instr[addr].op1.op    = ret_reg;
        instr[addr].op1.flags = OPF_DST;

        instr[addr].op2.type  = OP_BRTRG;
        instr[addr].op2.op    = btarg;
        instr[addr].op2.flags = OPF_SRC | OPF_ADDR;

        instr[addr].flags    |= FLAG_JSR;

        instr[addr].mnemonic  = mnm_jsr[int_mode];
    } else            /* J/JE/JD */
    {
        instr[addr].op1.type  = OP_BRTRG;
        instr[addr].op1.op    = btarg;
        instr[addr].op1.flags = OPF_SRC | OPF_ADDR;

        instr[addr].mnemonic  = mnm_jsr[int_mode + 4];
    }
}

/* ======================================================================== */
/*  DIS_IMPL_1OP_A -- decode HLT/SDBD/EIS/DIS                               */
/* ======================================================================== */
void dis_impl_1op_a(uint_32 addr)
{
    uint_32 w0, w1;

    /* -------------------------------------------------------------------- */
    /*  Make sure it's a valid opcode.                                      */
    /* -------------------------------------------------------------------- */
    if ((w0 = GET_WORD(addr)) > 3)
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  If this is a HLT instruction, mark it invalid if asked to do so.    */
    /* -------------------------------------------------------------------- */
    if (hlt_is_invalid && w0 == 0x000)
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  If this is an SDBD, make sure the instruction it modifies is an     */
    /*  indirect mode or immediate mode instruction that isn't MVO.         */
    /* -------------------------------------------------------------------- */
    if (w0 == 0x001)
    {
        w1 = GET_WORD(addr + 1);

        /* ---------------------------------------------------------------- */
        /*  The '(w1 & 0x3C0) < 0x280' makes sure it's MVI, ADD, SUB, CMP   */
        /*  AND or XOR.  The '(w1 & 0x038) < 0x080' makes sure it's not a   */
        /*  direct-mode instruction.                                        */
        /* ---------------------------------------------------------------- */
        if (NOT_CODE(addr + 1) || (w1 & 0x3C0) < 0x280 || (w1 & 0x038) < 0x080)
        {
            instr[addr].flags |= FLAG_INVOP;
            return;
        }
        /* ---------------------------------------------------------------- */
        /*  Go ahead and mark the next instruction as being modified by     */
        /*  SDBD.  We may revise this later once we discover br targets.    */
        /* ---------------------------------------------------------------- */
        instr[0xFFFF & (addr + 1)].flags |= FLAG_SDBD;
    }

    /* -------------------------------------------------------------------- */
    /*  Copy in the mnemonic and set the length.                            */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic = mnm_impl_1op_a[w0];
    instr[addr].len = 1;
}

/* ======================================================================== */
/*  DIS_IMPL_1OP_B -- decode TCI/CLRC/SETC                                  */
/* ======================================================================== */
void dis_impl_1op_b(uint_32 addr)
{
    uint_32 w0;

    /* -------------------------------------------------------------------- */
    /*  Nuke invalid ops, and TCI if we're treating rare ops as invalid.    */
    /* -------------------------------------------------------------------- */
    if ((w0 = GET_WORD(addr)) < 5 || w0 > 7 ||
        (rare_ops_are_invalid && w0 == 0x005))  /* TCI */
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Copy in the mnemonic and set the length.                            */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic = mnm_impl_1op_b[w0 - 4];
    instr[addr].len = 1;
}

/* ======================================================================== */
/*  DIS_NOP_SIN -- Decode NOP and SIN instructions.                         */
/* ======================================================================== */
void dis_nop_sin   (uint_32 addr)
{
    uint_32 w0;

    /* -------------------------------------------------------------------- */
    /*  If LSB != 0, treat it as invalid opcode no matter what.  Also, if   */
    /*  we're treating "rare as invalid", treat all SIN as invalid.         */
    /* -------------------------------------------------------------------- */
    w0 = GET_WORD(addr);

    if ((w0 & 1) ||
        (rare_ops_are_invalid && (w0 & 3) >= 2))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Copy in the mnemonic and set the length.                            */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic = w0 & 2 ? M_NOP : M_SIN;
    instr[addr].len = 1;
}

/* ======================================================================== */
/*  DIS_GSWD    -- Decode Get Status WorD instructions.                     */
/* ======================================================================== */
void dis_gwsd      (uint_32 addr)
{
    /* -------------------------------------------------------------------- */
    /*  Dirt simple.  One mnemonic.  Bits 0..1 are the destination.         */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic  = M_GSWD;
    instr[addr].op1.type  = OP_REG;
    instr[addr].op1.op    = GET_WORD(addr) & 3;
    instr[addr].op1.flags = OPF_DST;
    instr[addr].len = 1;
}

/* ======================================================================== */
/*  DIS_REG_1OP -- Decode single-operand register instructions.             */
/* ======================================================================== */
void dis_reg_1op   (uint_32 addr)
{
    uint_32 w0, op, dst;

    /* -------------------------------------------------------------------- */
    /*  Bits 0..2 are the destination (R0..R7).  Bits 3..5 are the opcode.  */
    /* -------------------------------------------------------------------- */
    w0 = GET_WORD(addr);
    op = (w0 >> 3) & 7;
    reg=  w0       & 7;

    /* -------------------------------------------------------------------- */
    /*  Opcode values 000 and 011 are invalid -- these alias other opcode   */
    /*  spaces.  We only check for them here for debug purposes.            */
    /* -------------------------------------------------------------------- */
    if (op == 0x000 || op = 0x006)
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Build the decoded instruction.                                      */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic  = mnm_reg_1op[op];
    instr[addr].op1.type  = OP_REG;
    instr[addr].op1.op    = dst;
    instr[addr].op1.flags = OPF_SRCDST;
    instr[addr].len = 1;
}

/* ======================================================================== */
/*  DIS_ROT_1OP -- Decode rotates and shifts.                               */
/* ======================================================================== */
void dis_rot_1op   (uint_32 addr)
{
    uint_32 w0, op, reg, amt; 

    /* -------------------------------------------------------------------- */
    /*  Bits 0..1 are the destination (R0 thru R3).  Bit 2 determines       */
    /*  whether we shift by 1 or by 2.  Bits 3..5 are the opcode.           */
    /* -------------------------------------------------------------------- */
    w0  = GET_WORD(addr);
    reg = (w0 >> 0) & 3;        /* R0 thru R3   */
    amt = (w0 >> 2) & 1;        /* ,1 or ,2     */
    op  = (w0 >> 3) & 7;        /* which opcode */

    /* -------------------------------------------------------------------- */
    /*  Build the instruction.                                              */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic  = mnm_rot_1op[op];

    instr[addr].op1.type  = OP_REG;
    instr[addr].op1.op    = reg;
    instr[addr].op1.flags = OPF_SRCDST;

    instr[addr].op2.type  = OP_IMMED;
    instr[addr].op2.op    = amt + 1;
    instr[addr].op2.flags = OPF_SRC;

    instr[addr].len = 1;
}


/* ======================================================================== */
/*  DIS_REG_2OP -- Decode MOVR/ADDR/SUBR/CMPR/ANDR/XORR.                    */
/* ======================================================================== */
void dis_reg_2op   (uint_32 addr)
{
    uint_32 w0, src, dst, op;
    mnm_t m;
    
    /* -------------------------------------------------------------------- */
    /*  Bits 0..2 are the source, bits 3..5 are the dest, and bits 6..8     */
    /*  are the opcode.                                                     */
    /* -------------------------------------------------------------------- */
    w0  = GET_WORD(addr);
    src = (w0 >> 0) & 7;
    dst = (w0 >> 3) & 7;
    op  = (w0 >> 6) & 7;

    /* -------------------------------------------------------------------- */
    /*  Opcodes 000 and 001 are illegal -- they alias other opcode spaces.  */
    /* -------------------------------------------------------------------- */
    if (op < 2)
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Build the instruction.                                              */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic  = m = mnm_reg_1op[op];
    
    instr[addr].op1.type  = OP_REG;
    instr[addr].op1.op    = src;
    instr[addr].op1.flags = OPF_SRC;

    instr[addr].op2.type  = OP_REG;
    instr[addr].op2.op    = dst;
    instr[addr].op2.flags = m == M_MOVR ? OPF_DST :
                            m == M_CMPR ? OPF_SRC : OPF_SRCDST;

    instr[addr].len      = 1;
}

/* ======================================================================== */
/*  DIS_COND_BR -- Decode all "internal" conditional branches, and NOPP.    */
/* ======================================================================== */
void dis_cond_br   (uint_32 addr)
{
    uint_32 w0, cnd, ofs, dir, trg, z;

    /* -------------------------------------------------------------------- */
    /*  Make sure target word is in implemented memory.                     */
    /* -------------------------------------------------------------------- */
    if (IS_EMPTY(addr + 1))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Bits 0..3 are the condition code.  Bit 5 determines branch target   */
    /*  direction.  Displacement is in second word.                         */
    /* -------------------------------------------------------------------- */
    w0  = GET_WORD(addr);
    cnd = (w0 >> 0) & 0xF;
    dir = (w0 >> 5) & 1;
    ofs = GET_WORD(addr + 1);
    trg = addr + 2 + (dir ? ~ofs : ofs);

    /* -------------------------------------------------------------------- */
    /*  Make sure NOPP has a 'zero' for its offset.  Both 'dir' and 'ofs'   */
    /*  must be zero for NOPP to be valid, as far as the disassembler is    */
    /*  concerned.                                                          */
    /* -------------------------------------------------------------------- */
    if (cnd == 0x8 && (dir != 0 || ofs != 0))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Check to make sure branch target doesn't wrap, unless user has      */
    /*  explicitly allowed it.                                              */
    /* -------------------------------------------------------------------- */
    if (cnd != 0x8 && !allow_branch_target_wrap && (trg & 0xFFFF0000))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    trg &= 0xFFFF;

    /* -------------------------------------------------------------------- */
    /*  Build the instruction.                                              */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic      = mnm_cond_br[op];
    
    if (cnd != 0x8)
    {
        instr[addr].op1.type  = OP_BRTRG;
        instr[addr].op1.op    = trg;
        instr[addr].op1.flags = OPF_SRC | OPF_ADDR;

        instr[addr].flags    |= FLAG_BRANCH;
        instr[addr].br_target = trg;

        /* ---------------------------------------------------------------- */
        /*  All branches except "B" are conditional branches.               */
        /* ---------------------------------------------------------------- */
        if (cnd != 0x0)
            instr[addr].flags |= FLAG_CONDBR;
    }

    instr[addr].len      = 2;
}

/* ======================================================================== */
/*  DIS_BEXT    -- Decode external-condition branches.                      */
/* ======================================================================== */
void dis_bext      (uint_32 addr)
{
    uint_32 w0, cnd, ofs, dir, trg, z;

    /* -------------------------------------------------------------------- */
    /*  Treat BEXT as 'rare' opcode, since it shouldn't appear in any       */
    /*  Intellivision program.                                              */
    /* -------------------------------------------------------------------- */
    if (rare_ops_are_invalid || IS_EMPTY(addr + 1))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Make sure target word is in implemented memory.                     */
    /* -------------------------------------------------------------------- */
    if (IS_EMPTY(addr + 1))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Bits 0..3 are the condition code.  Bit 5 determines branch target   */
    /*  direction.  Displacement is in second word.                         */
    /* -------------------------------------------------------------------- */
    w0  = GET_WORD(addr);
    cnd = (w0 >> 0) & 0xF;
    dir = (w0 >> 5) & 1;
    ofs = GET_WORD(addr + 1);
    trg = addr + 2 + (dir ? ~ofs : ofs);

    /* -------------------------------------------------------------------- */
    /*  Check to make sure branch target doesn't wrap, unless user has      */
    /*  explicitly allowed it.                                              */
    /* -------------------------------------------------------------------- */
    if (!allow_branch_target_wrap && (trg & 0xFFFF0000))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }
    trg &= 0xFFFF;

    /* -------------------------------------------------------------------- */
    /*  Build the instruction.                                              */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic  = M_BEXT;
    
    instr[addr].op1.type  = OP_IMMED;
    instr[addr].op1.op    = cnd;
    instr[addr].op1.flags = OPF_SRC;

    instr[addr].op2.type  = OP_BRTRG;
    instr[addr].op2.op    = trg;
    instr[addr].op2.flags = OPF_SRC | OPF_ADDR;

    instr[addr].flags    |= FLAG_BRANCH | FLAG_CONDBR;
    instr[addr].br_target = trg;

    instr[addr].len      = 2;
}

/* ======================================================================== */
/*  DIS_IMM_2OP -- Decode immediate mode instructions                       */
/* ======================================================================== */
void dis_imm_2op   (uint_32 addr)
{
    uint_32 w0, w1, w2, op, dst, imm;
    mnm_t   m;
    int     is_sdbd = 0;

    /* -------------------------------------------------------------------- */
    /*  Destination is in bits 0..2.  Opcode is in bits 6..8.               */
    /* -------------------------------------------------------------------- */
    w0  = GET_WORD(addr);
    op  = (w0 >> 6) & 7;
    dst = (w0 >> 0) & 7;

    /* -------------------------------------------------------------------- */
    /*  Chuck opcode 000, as it's illegal.  Also, if opcode is MVOI (001)   */
    /*  and we treat 'rare' as 'illegal', then chuck it as well.            */
    /* -------------------------------------------------------------------- */
    if (op == 0 || (rare_ops_are_illegal && op == 1))
    {
        instr[addr].flags |= FLAG_INVOP;
        if (instr[addr].flags & FLAG_SDBD)
            instr[0xFFFF & (addr - 1)].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Read our immediate constant.                                        */
    /*                                                                      */
    /*  We examine our own "FLAG_SDBD" to decide if our immediate constant  */
    /*  spans 1 or 2 words.  If FLAG_SDBD is set, but we fail validity      */
    /*  tests, we clear FLAG_SDBD and mark the SDBD as invalid code.        */
    /* -------------------------------------------------------------------- */
    w1  = GET_WORD(addr + 1);
    w2  = GET_WORD(addr + 2);
    imm = w1;    /* default is that word 1 is the immediate operand */

    if (IS_EMPTY(addr + 1))
    {
        instr[addr].flags |= FLAG_INVOP;
        if (instr[addr].flags & FLAG_SDBD)
            instr[0xFFFF & (addr - 1)].flags |= FLAG_INVOP;
        return;
    }

    if (instr[addr].flags & FLAG_SDBD)
    {
        /* ---------------------------------------------------------------- */
        /*  SDBD is valid only if:                                          */
        /*   -- second word is in implemented memory.                       */
        /*   -- both words are empty in bits 8..15.                         */
        /*   -- instruction is not MVOI (opcode 001).                       */
        /* ---------------------------------------------------------------- */
        if (op == 1 || IS_EMPTY(addr + 2) || (w1 & 0xFF00) || (w2 & 0xFF00))
        {
            instr[0xFFFF & (addr - 1)].flags |= FLAG_INVOP;
            instr[addr].flags &= ~FLAG_SDBD;
        } else
        {
            imm = (w1 & 0x00FF) | ((w2 << 8) & 0xFF00);
            is_sdbd = 1;
        }
    }

    /* -------------------------------------------------------------------- */
    /*  Build the instruction.                                              */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic = m = mnm_imm_2op[op];
    
    if (m == M_MVOI)
    {
        instr[addr].op1.type  = OP_REG;
        instr[addr].op1.op    = dst;
        instr[addr].op1.flags = OPF_SRC;
    
        instr[addr].op2.type  = OP_IMMED;
        instr[addr].op2.op    = imm;
        instr[addr].op2.flags = OPF_DST;
    } else
    {
        instr[addr].op1.type  = OP_IMMED;
        instr[addr].op1.op    = imm;
        instr[addr].op1.flags = OPF_SRC;
    
        instr[addr].op2.type  = OP_REG;
        instr[addr].op2.op    = dst;
        instr[addr].op2.flags = m == M_MVII ? OPF_DST :
                                m == M_CMPI ? OPF_SRC : OPF_SRCDST;
    }

    instr[addr].len = 2 + is_sdbd;
}

/* ======================================================================== */
/*  DIS_DIR_2OP -- Decode direct mode instructions                          */
/* ======================================================================== */
void dis_dir_2op   (uint_32 addr)
{
    uint_32 w0, w1, w2, op, dst;
    mnm_t   m;

    /* -------------------------------------------------------------------- */
    /*  Destination is in bits 0..2.  Opcode is in bits 6..8.               */
    /* -------------------------------------------------------------------- */
    w0  = GET_WORD(addr);
    w1  = GET_WORD(addr + 1); /* direct address */
    op  = (w0 >> 6) & 7;
    dst = (w0 >> 0) & 7;

    /* -------------------------------------------------------------------- */
    /*  SDBD is always illegal for direct-mode.                             */
    /* -------------------------------------------------------------------- */
    if (instr[addr].flags & FLAG_SDBD)
        instr[0xFFFF & (addr - 1)].flags |= FLAG_INVOP;

    /* -------------------------------------------------------------------- */
    /*  Chuck opcode 000, as it's illegal.                                  */
    /*  Also, make sure our direct address is legal.                        */
    /* -------------------------------------------------------------------- */
    if (op == 0 || IS_EMPTY(addr + 1))
    {
        instr[addr].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  Build the instruction.                                              */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic = m = mnm_ind_2op[op];
    
    if (m == M_MVO)  /* MVO */
    {
        instr[addr].op1.type  = OP_REG;
        instr[addr].op1.op    = dst;
        instr[addr].op1.flags = OPF_SRC;
    
        instr[addr].op2.type  = OP_DIRADDR;
        instr[addr].op2.op    = w1;
        instr[addr].op2.flags = OPF_DST | OPF_ADDR;
    } else
    {
        instr[addr].op1.type  = OP_DIRADDR;
        instr[addr].op1.op    = w1;
        instr[addr].op1.flags = OPF_SRC | OPF_ADDR;
    
        instr[addr].op2.type  = OP_REG;
        instr[addr].op2.op    = dst;
        instr[addr].op2.flags = m == M_MVI ? OPF_DST :
                                m == M_CMP ? OPF_SRC : OPF_SRCDST;
    }

    instr[addr].len = 1;
}

/* ======================================================================== */
/*  DIS_IND_2OP -- Decode indirect mode instructions                        */
/* ======================================================================== */
void dis_ind_2op   (uint_32 addr)
{
    uint_32 w0, op, op1, op2;
    mnm_t   m;

    /* -------------------------------------------------------------------- */
    /*  Destination is in bits 0..2.  Source is in bits 3..5.  Opcode is    */
    /*  in bits 6..8.                                                       */
    /* -------------------------------------------------------------------- */
    w0  = GET_WORD(addr);
    op  = (w0 >> 6) & 7;
    op1 = (w0 >> 3) & 7;
    op2 = (w0 >> 0) & 7;

    /* -------------------------------------------------------------------- */
    /*  Chuck opcode 000, as it's illegal.                                  */
    /* -------------------------------------------------------------------- */
    if (op == 0)
    {
        instr[addr].flags |= FLAG_INVOP;
        if (instr[addr].flags & FLAG_SDBD)
            instr[0xFFFF & (addr - 1)].flags |= FLAG_INVOP;
        return;
    }

    /* -------------------------------------------------------------------- */
    /*  If FLAG_SDBD is set but we're MVO@, or we're reading via R6, nuke   */
    /*  FLAG_SDBD.  (SDBD is incompatible w/ MVO@ and stack accesses.)      */
    /* -------------------------------------------------------------------- */
    if ((instr[addr].flags & FLAG_SDBD) && (op == 1 || op1 == 6))
    {
        instr[0xFFFF & (addr - 1)].flags |= FLAG_INVOP;
        instr[addr].flags &= ~FLAG_SDBD;
    }

    /* -------------------------------------------------------------------- */
    /*  Build the instruction.                                              */
    /* -------------------------------------------------------------------- */
    instr[addr].mnemonic = m = mnm_ind_2op[op];
    
    if (m == M_MVO_)
    {
        instr[addr].op1.type  = OP_REG;
        instr[addr].op1.op    = op1;
        instr[addr].op1.flags = OPF_SRC;
    
        instr[addr].op2.type  = OP_REG;
        instr[addr].op2.op    = op2;
        instr[addr].op2.flags = OPF_DST | OPF_IND;
    } else
    {
        instr[addr].op1.type  = OP_REG;
        instr[addr].op1.op    = op1;
        instr[addr].op1.flags = OPF_SRC | OPF_IND;
    
        instr[addr].op2.type  = OP_REG;
        instr[addr].op2.op    = op2;
        instr[addr].op2.flags = m == M_MVI_ ? OPF_DST :
                                m == M_CMP_ ? OPF_SRC : OPF_SRCDST;
    }

    instr[addr].len = 1;
}

/* ======================================================================== */
/*  DECODE_INSTRS   -- Try to decode all instructions.  Mark obviously      */
/*                     invalid instructions as such.  If enabled, mark      */
/*                     "rare" instructions illegal too.                     */
/* ======================================================================== */
void decode_instrs(int addr_lo, int addr_hi)
{
    uint_32 addr, w0;
    int sdbd_ok = 0;
    
    /* -------------------------------------------------------------------- */
    /*  Step by single addresses within range.  Although some instructions  */
    /*  are multi-word, we don't know where instruction boundaries are yet. */
    /*  Just decode all words as if they are instructions and then sort it  */
    /*  out more fully later.                                               */
    /* -------------------------------------------------------------------- */
    for (addr = addr_lo; addr < addr_hi; addr++)
    {
        /* ---------------------------------------------------------------- */
        /*  Instructions that set bits 10..15 are invalid.  Skip them.      */
        /* ---------------------------------------------------------------- */
        if (GET_WORD(addr) & 0xFC00)
        {
            instr[addr].flags |= FLAG_INVOP;
            continue;
        }

        /* ---------------------------------------------------------------- */
        /*  Skip over addresses we've previously identified as not code.    */
        /*  Examples are explicit data, and invalid opcodes.                */
        /* ---------------------------------------------------------------- */
        if (NOT_CODE(addr))
            continue;

        w0 = GET_WORD(addr);
        assert((w0 & 0xFC00) == 0);


        /* ---------------------------------------------------------------- */
        /*  Pull apart the opcode by opcode space.                          */
        /*                                                                  */
        /*                                                                  */
        /*      00 0000 0100    Absolute jump instructions                  */
        /*      00 0000 00oo    Implied 1-op arithmetic  (a)                */
        /*      00 0000 01oo    Implied 1-op arithmetic  (b)                */
        /*      00 0011 01oo    NOP, SIN instructions                       */
        /*      00 0011 00oo    GSWD -- Get Status WorD insn.               */
        /*      00 00oo oddd    Combined Src/Dst Register 1-op              */
        /*      00 01oo omrr    Rotate/Shift Register 1-op                  */
        /*      0o ooss sddd    Register  -> Register 2-op arith            */
        /*      10 00z0 cccc    Conditional branch (internal condition)     */
        /*      10 00z1 cccc    Conditional branch (external condition)     */
        /*      1o oo11 1ddd    Immediate -> Register 2-op arith            */
        /*      1o oo00 0ddd    Direct    -> Register 2-op arith            */
        /*      1o oomm mddd    Indirect  -> Register 2-op arith            */
        /*                                                                  */
        /* ---------------------------------------------------------------- */
        if      ((w0 & 0x3FF)==0x004) dis_jump      (addr); /* 00 0000 0100 */
        else if ((w0 & 0x3FC)==0x000) dis_impl_1op_a(addr); /* 00 0000 00oo */
        else if ((w0 & 0x3FC)==0x004) dis_impl_1op_b(addr); /* 00 0000 01oo */
        else if ((w0 & 0x3FC)==0x034) dis_nop_sin   (addr); /* 00 0011 01oo */
        else if ((w0 & 0x3FC)==0x030) dis_gwsd      (addr); /* 00 0011 00oo */
        else if ((w0 & 0x3C0)==0x000) dis_reg_1op   (addr); /* 00 00oo oddd */
        else if ((w0 & 0x3C0)==0x040) dis_rot_1op   (addr); /* 00 01oo omrr */
        else if ((w0 & 0x200)==0x000) dis_reg_2op   (addr); /* 0o ooss sddd */
        else if ((w0 & 0x3D0)==0x200) dis_cond_br   (addr); /* 10 00z0 cccc */
        else if ((w0 & 0x3D0)==0x210) dis_bext      (addr); /* 10 00z1 cccc */
        else if ((w0 & 0x238)==0x238) dis_imm_2op   (addr); /* 1o oo11 1ddd */
        else if ((w0 & 0x238)==0x200) dis_dir_2op   (addr); /* 1o oo00 0ddd */
        else                          dis_ind_2op   (addr); /* 1o oomm mddd */

    }
}

/* ======================================================================== */
/*  MARK_EMPTY -- Mark an address range as not holding "local code."        */
/* ======================================================================== */
void mark_empty(int addr_lo, int addr_hi)
{
    int addr;

    for (addr = addr_lo; addr <= addr_hi; addr++)
        instr[addr].flags |= FLAG_EMPTY;
}

/* ======================================================================== */
/*  FIND_FUNKY_BRANCHES() -- Look for arithmetic on R7 and convert these    */
/*                           instructions to branches, *or* treat them as   */
/*                           invalid if they look invalid.                  */
/* ======================================================================== */
int find_funky_branches()
{
    uint_32 addr;
    int changed = 0;

    /* -------------------------------------------------------------------- */
    /*  Examine all valid decoded instructions, and see if any of them      */
    /*  operate directly on the program counter as a destination.  Some     */
    /*  of these are highly suspicious and can be converted into invalid    */
    /*  opcodes.  Other instructions are useful -- INCR PC, DECR PC, ADCR   */
    /*  PC and ADDR PC are all common tricks.                               */
    /* -------------------------------------------------------------------- */
    for (addr = 0; addr <= 0xFFFF; addr++)
    {
        /* ---------------------------------------------------------------- */
        /*  Skip code that's already identified as 'bad'.                   */
        /*  Skip code that's already identified as branches.                */
        /* ---------------------------------------------------------------- */
        if (NOT_CODE(addr) || IS_BRANCH(addr))
            continue;

        /* ---------------------------------------------------------------- */
        /*  Check the first operand.  Only 1op Reg format can set this to   */
        /*  "R7 is destination".                                            */
        /* ---------------------------------------------------------------- */
        if ( instr[addr].op1.type == OP_REG &&
             instr[addr].op1.op   == 7      &&
            (instr[addr].op1.flags & OPF_DST))
        {
            switch (instr[addr].mnemonic)
            {
                /* -------------------------------------------------------- */
                /*  ADCR looks like a short forward conditional branch.     */
                /* -------------------------------------------------------- */
                case M_ADCR: 
                {
                    instr[addr].br_target = addr + 2;
                    instr[addr].flags    |= FLAG_BRANCH | FLAG_CONDBR;
                    break;
                }
                    
                /* -------------------------------------------------------- */
                /*  INCR/DECR are unconditional branches.                   */
                /* -------------------------------------------------------- */
                case M_INCR:
                {
                    instr[addr].br_target = addr + 2;
                    instr[addr].flags    |= FLAG_BRANCH;
                    break;
                }

                case M_DECR:
                {
                    instr[addr].br_target = addr;
                    instr[addr].flags    |= FLAG_BRANCH;
                    break;
                }

                case M_COMR:
                {
                    if (suspicious_pc_math_is_invalid)
                    {
                        instr[addr].flags |= FL_INVOP;
                        changed++;
                    } else
                    {
                        instr[addr].br_target = addr ^ 0xFFFF;
                        instr[addr].flags    |= FLAG_BRANCH;
                    }
                    break;
                }

                case M_NEGR:
                {
                    if (suspicious_pc_math_is_invalid)
                    {
                        instr[addr].flags |= FL_INVOP;
                        changed++;
                    } else
                    {
                        instr[addr].br_target = (-addr) & 0xFFFF;
                        instr[addr].flags    |= FLAG_BRANCH;
                    }
                    break;
                }

                case M_RSWD:    /* Treat RSWD R7 as always illegal, period */
                {
                    instr[addr].flags |= FL_INVOP;
                    changed++;
                    break;
                }
                default:        /* Uhoh?! */
                {
                    fprintf(stderr, 
                            "Instruction '%s' at $%.4X sets op1 as dst R7\n",
                            mnemonic[instr[addr].mnemonic], addr);
                    break;
                }
            }

            continue;
        }
                    
        /* ---------------------------------------------------------------- */
        /*  Check the second operand.                                       */
        /* ---------------------------------------------------------------- */
        if ( instr[addr].op2.type == OP_REG &&
             instr[addr].op2.op   == 7      &&
            (instr[addr].op2.flags & OPF_DST))
        {
            /* ------------------------------------------------------------ */
            /*  In general, we permit MOV, ADD, SUB.  Anything else looks   */
            /*  pretty darn suspicious.  We can calculate branch targets    */
            /*  for ADDI, SUBI, MVII, but nothing else for now.             */
            /*                                                              */
            /*  Also, ADDI/SUBI that result in targets outside the 16-bit   */
            /*  range are suspicious.                                       */
            /* ------------------------------------------------------------ */

            switch (instr[addr].mnemonic)
            {
                /* -------------------------------------------------------- */
                /*  MVII is basically identical to J, only shorter.         */
                /*  In a 16-bit ROM, B is a better choice though.           */
                /* -------------------------------------------------------- */
                case M_MVII:
                {
                    instr[addr].br_target = instr[addr].op1.op;
                    instr[addr].flags    |= FLAG_BRANCH;
                    break;
                }

                /* -------------------------------------------------------- */
                /*  These are still somewhat suspicious -- why not use "B"? */
                /* -------------------------------------------------------- */
                case M_ADDI:
                case M_SUBI:
                {
                    uint_32 targ;
                    
                    if (instr[addr].mnemonic == M_ADDI)
                    {
                        targ = addr + instr[addr].len + instr[addr].op1.op;
                    } else
                    {
                        targ = addr + instr[addr].len - instr[addr].op1.op;
                    }

                    if (suspicious_pc_math_is_invalid && addr > 0xFFFF)
                    {
                        instr[addr].flags |= FLAG_INVOP;
                        changed++;
                        break;
                    }

                    instr[addr].br_target = addr & 0xFFFF;
                    instr[addr].flags    |= FLAG_BRANCH;
                    break;
                }

                /* -------------------------------------------------------- */
                /*  These are very suspicious.                              */
                /* -------------------------------------------------------- */
                case M_ANDI:
                case M_XORI:
                {
                    uint_32 targ;

                    if (suspicious_pc_math_is_invalid)
                    {
                        instr[addr].flags |= FLAG_INVOP;
                        changed++;
                        break;
                    }
                    
                    if (instr[addr].mnemonic == M_ANDI)
                    {
                        targ = (addr + instr[addr].len) & instr[addr].op1.op;
                    } else
                    {
                        targ = (addr + instr[addr].len) ^ instr[addr].op1.op;
                    }

                    instr[addr].br_target = addr & 0xFFFF;
                    instr[addr].flags    |= FLAG_BRANCH;
                    break;
                }


                /* -------------------------------------------------------- */
                /*  These are fairly common, especially MVI@, MOVR, ADDR.   */
                /* -------------------------------------------------------- */
                case M_MVI:  case M_ADD:  case M_SUB:
                case M_MVI_: case M_ADD_: case M_SUB_:
                case M_MOVR: case M_ADDR: case M_SUBR:
                {
                    instr[addr].br_target = BTARG_UNK;
                    instr[addr].flags    |= FLAG_BRANCH;
                    break;
                }

                /* -------------------------------------------------------- */
                /*  Note that I shouldn't see MVO or CMP.  This is for      */
                /*  error checking purposes only.                           */
                /* -------------------------------------------------------- */
                case M_MVO:  case M_CMP:
                case M_MVO_: case M_CMP_:
                case M_MVOI: case M_CMPI:
                {
                    fprintf(stderr, 
                            "WARNING: '%s' at $%.4x appears to write R7?\n",
                            mnemonic[instr[addr].mnemonic], addr);
                    break;
                }

                /* -------------------------------------------------------- */
                /*  Treat everything else as invalid/unknown target.        */
                /* -------------------------------------------------------- */
                default:
                {
                    if (suspicious_pc_math_is_invalid)
                    {
                        instr[addr].flags |= FLAG_INVOP;
                        changed++;
                    } else
                    {
                        instr[addr].br_target = BTARG_UNK;
                        instr[addr].flags    |= FLAG_BRANCH;
                    }
                    break;
                }
            }
            continue;
        }
    }

    return changed;
}

/* ======================================================================== */
/*  PROPAGATE_INVOP  -- Scan code backwards, propagating "INVOP" attrs      */
/*                      back towards unconditional control transfers.       */
/*                      The idea is that you can't reach an invalid op by   */
/*                      "falling through" from a valid op.                  */
/* ======================================================================== */
int propagate_invop(void)
{
    int_32 addr, step;
    int changed = 0;

    /* -------------------------------------------------------------------- */
    /*  This is tricky.  We scan for "truly invalid" instructions.  To      */
    /*  find these, though, during a *backwards* scan is tricky.  Why?      */
    /*  Instructions can be up to 3 words long.  In a 3-word-long instr,    */
    /*  the 2nd and 3rd words might look invalid even though the 3-word     */
    /*  combo is perfectly valid.  Thus, when we step backwards looking     */
    /*  for starting points, we do a 'greedy gobble', looking for the       */
    /*  longest instruction that ends at our current position, and step to  */
    /*  there.                                                              */
    /* -------------------------------------------------------------------- */

    /*  Find first good instruction.  */
    for (addr = 0xFFFF; addr > 0x0002; addr--)
    {
        if (!IS_EMPTY(addr) && MAYBE_CODE(addr))
            break;
    }

    for (; addr > 0x0002; addr -= step)
    {
        /* ---------------------------------------------------------------- */
        /*  Only analyse ROM we know about.                                 */
        /* ---------------------------------------------------------------- */
        if (IS_EMPTY(addr))
            continue;

        /* ---------------------------------------------------------------- */
        /*  If this word is 'ok', try stepping to a previous instruction.   */
        /* ---------------------------------------------------------------- */
        if (MAYBE_CODE(addr))
        {
            if      (MAYBE_CODE(addr - 3) && instr[addr].len == 3) step = 3;
            else if (MAYBE_CODE(addr - 2) && instr[addr].len >= 2) step = 2;
            else                                                   step = 1;
            continue;
        }

        /* ---------------------------------------------------------------- */
        /*  Ok, we're at an invalid instruction.  Now chase this to         */
        /*  earlier addresses until we hit an unconditional branch.         */
        /* ---------------------------------------------------------------- */
        step = 0;
        while (addr > 2 && !IS_EMPTY(addr))
        {
            /* len 1 branches are "INCR PC", "DECR PC", etc. */
            if (MAYBE_CODE(addr - 1) && 
                IS_BRANCH (addr - 1) &&
                instr[addr - 1].len == 1)
                break;

            /* len 2 branches are "B", "Bcond", and some PC arith. */
            if (MAYBE_CODE(addr - 2) && 
                IS_BRANCH (addr - 2) && 
                !IS_CONDBR(addr - 2) &&
                instr[addr - 2].len == 2)
                break;

            /* len 3 branches are J/JSR, and some PC arith.  */
            if (MAYBE_CODE(addr - 3) && 
                IS_BRANCH (addr - 3) &&
                instr[addr - 3].len == 3)
                break;

            if ((instr[addr].flags & FLAG_INVOP) == 0)
                changed++;

            instr[addr].flags |= FLAG_INVOP;
            instr[addr].flags &= (FLAG_BRANCH|FLAG_CONDBR|FLAG_JSR);

            addr--;
        }
    }

    return changed;
}


/* ======================================================================== */
/*  KILL_BAD_BRANCHES -- Look for branches to invalid opcodes, and mark     */
/*                       those branches as invalid instructions.            */
/* ======================================================================== */
int kill_bad_branches(void)
{
    uint_32 addr, targ;
    int changed = 0;

    for (addr = 0; addr <= 0xFFFF; addr++)
    {
        /* ---------------------------------------------------------------- */
        /*  Skip already-invalid instructions.                              */
        /* ---------------------------------------------------------------- */
        if (NOT_CODE(addr))
            continue;

        /* ---------------------------------------------------------------- */
        /*  See if this is a branch to a known target.                      */
        /* ---------------------------------------------------------------- */
        if ((instr[addr].flags & FLAG_BRANCH) &&
            (targ = instr[addr].br_target) <= 0xFFFF)
        {
            /* ------------------------------------------------------------ */
            /*  If so, see if it's a branch into ROM we know about.         */
            /* ------------------------------------------------------------ */
            if (IS_EMPTY(instr[addr]))
                continue;

            /* ------------------------------------------------------------ */
            /*  If the target is known, but not valid code, mark us inval.  */
            /* ------------------------------------------------------------ */
            if (NOT_CODE(targ))
            {
                instr[addr].flags |= FLAG_INVOP;
                instr[addr].flags &= (FLAG_BRANCH|FLAG_CONDBR|FLAG_JSR);
                changed++;
            }
        }
    }

    return changed;
}

/* ======================================================================== */
/*  BRTRG_VS_SDBD    -- Scan for branches that target instructions w/       */
/*                      FLAG_SDBD set.  Decide whether branch or instr      */
/*                      needs changing.                                     */
/* ======================================================================== */
int brtrg_vs_sdbd(void)
{
    uint_32 addr, targ;
    int need_rescan;
    int changed = 0;

    /* -------------------------------------------------------------------- */
    /*  For each potential branch target, do the following:                 */
    /*                                                                      */
    /*   -- If target's exact address is unknown, skip it.                  */
    /*   -- If the target is not modified by SDBD, ignore it.               */
    /*   -- If the target is modified by SDBD, but is indirect mode,        */
    /*      remove FLAG_SDBD and move on.                                   */
    /*   -- If the target is an immediate-mode instruction modified by      */
    /*      SDBD:                                                           */
    /*       -- If its third word is INVALID and the target is itself       */
    /*          not a "branch" (eg. ADDI #foo, PC), mark this branch        */
    /*          as an invalid instruction.                                  */
    /*       -- Otherwise, remove FLAG_SDBD from instruction and redecode.  */
    /* -------------------------------------------------------------------- */

    for (addr = 0; addr <= 0xFFFF; addr++)
    {
        /* ---------------------------------------------------------------- */
        /*  Only handle stuff that still looks like code.                   */
        /* ---------------------------------------------------------------- */
        if (NOT_CODE(addr))
            continue;

        if (IS_BRANCH(addr))
        {
            targ = instr[addr].br_target;

            /* ------------------------------------------------------------ */
            /*  Only deal w/ known targets to SDBD-affected instructions.   */
            /* ------------------------------------------------------------ */
            if (targ > 0xFFFF ||                         /* unknown target  */
                (instr[targ].flags & FLAG_SDBD) == 0)    /* not SDBD target */
                continue;

            /* ------------------------------------------------------------ */
            /*  Handle direct mode by just nuking 'SDBD' flag.              */
            /* ------------------------------------------------------------ */
            if (instr[targ].op1.type == OP_REG)          /* indirect mode   */
            {
                instr[targ].flags &= ~FLAG_SDBD;
                /* don't count this as 'changed' though. */
                continue;
            }

            /* ------------------------------------------------------------ */
            /*  The other case we handle is immediate-mode, so skip out on  */
            /*  anything else 'weird'.                                      */
            /* ------------------------------------------------------------ */
            if (instr[targ].op1.type != OP_IMMED)       /* not immedate     */
                continue;

            /* ------------------------------------------------------------ */
            /*  Ok, if we remove SDBD, do we expose bad code?               */
            /* ------------------------------------------------------------ */
            if (NOT_CODE(targ + 2) && !IS_BRANCH(targ))
            {
                /* -------------------------------------------------------- */
                /*  Yes:  Mark the branch invalid.                          */
                /* -------------------------------------------------------- */
                instr[addr].flags |= FLAG_INVOP;
                instr[addr].flags &= ~(FLAG_BRANCH|FLAG_CONDBR|FLAG_JSR);
                changed++;
            } else
            {
                /* -------------------------------------------------------- */
                /*  No:  Remove SDBD from the target.                       */
                /* -------------------------------------------------------- */
                instr[targ].flags &= ~(FLAG_SDBD);
                changed++;
            }
        }
    }
                
    return changed;        
}

/* ======================================================================== */
/*  GENERATE_LABELS  -- Scan for all hard memory reference and generate /   */
/*                      look up labels for them.                            */
/* ======================================================================== */
void generate_labels(void)
{
    int addr; 
    uint_32 targ;
    char *lbl;

    /* -------------------------------------------------------------------- */
    /*  Scan for *all* memory references.  References that point within     */
    /*  our readable, preloaded pages are "local".  All others are          */
    /*  considered "global".                                                */
    /* -------------------------------------------------------------------- */
    for (addr = 0; addr <= 0xFFFF; addr++)
    {
        if (NOT_CODE(addr))
            continue;

        /* ---------------------------------------------------------------- */
        /*  Instructions can have only one address operand.  We look at     */
        /*  the two operands first, and then at the branch target (if this  */
        /*  is marked as a branch).                                         */
        /* ---------------------------------------------------------------- */
        if      (instr[addr].op1.flags & OPF_ADDR) targ=instr[addr].op1.op;
        else if (instr[addr].op2.flags & OPF_ADDR) targ=instr[addr].op2.op;
        else if (IS_BRANCH(addr) && 
                 instr[addr].br_target <= 0xFFFF)  targ=instr[addr].br_target;
        else                                       continue;

        /* ---------------------------------------------------------------- */
        /*  Avoid oddball occurrences.                                      */
        /* ---------------------------------------------------------------- */
        if (targ < 0 || targ > 0xFFFF)
            continue;

        /* ---------------------------------------------------------------- */
        /*  Look up the label.  If one doesn't exist, symtab() will make    */
        /*  one up for us with the prefix we specify.                       */
        /* ---------------------------------------------------------------- */
        lbl  = symtab_getsym(symtab, targ, IS_EMPTY(targ) ? 'G' : 'L', 0);
        if (lbl)
            instr[targ].label = strdup(lbl);
    }
}

/* ======================================================================== */
/*  DO_DISASM -- Top level driver that disassembles everything              */
/* ======================================================================== */
void do_disasm(void)
{
    int addr, p, sa = -1, ea = -1;
    int tot_changes = 0, changes, c;
    
    /* -------------------------------------------------------------------- */
    /*  Scan for preload pages that are also marked readable and not        */
    /*  bankswitched.                                                       */
    /* -------------------------------------------------------------------- */
    VB_PRINTF(0, ("Initial decode pass...\n"));

    for (addr = 0; addr <= 0xFF00; addr += 256)
    {
        p = addr >> 8;

        if ((GET_BIT(icart.preload,  p) != 0) &&
            (GET_BIT(icart.readable, p) != 0) &&
            (GET_BIT(icart.banksw,   p) == 0))
        {
            if (sa == -1) 
                sa = addr;

            ea = addr + 255;

            decode_instr(addr, addr + 255);
        } else
        {
            mark_empty(addr, addr + 255);
            if (sa >= 0)
                VB_PRINTF(1, (">   $%.4X - $%.4X\n", sa, ea));

            sa = ea = -1;
        }
    }

    if (sa >= 0)
        VB_PRINTF(1, (">   $%.4X - $%.4X\n", sa, ea));

    sa = ea = -1;

    /* -------------------------------------------------------------------- */
    /*  Now, if enabled, do various "analysis passes".                      */
    /* -------------------------------------------------------------------- */
    if (!skip_basic_analysis)
    {
        VB_PRINTF(0, ("Performing basic analysis.\n"));
        VB_PRINTF(1, ("> Identifying arithmetic branches...\n"));
        tot_changes = find_funky_branches();
   
        do
        {
            changes = 0;
   
            VB_PRINTF(1, ("> Propagating invalid opcodes...\n"));
            changes += c = propagate_invop();
            VB_PRINTF(2, (">>  %d words marked invalid\n", c));
   
            VB_PRINTF(1, ("> Marking invalid branches...\n"));
            changes += c = kill_bad_branches();
            VB_PRINTF(2, (">>  %d words marked invalid\n", c));
        
            VB_PRINTF(1, ("> Analysing SDBD/branch interaction...\n"));
            changes += c = brtrg_vs_sdbd();
            VB_PRINTF(2, (">>  %d words marked invalid\n", c));
   
            tot_changes += 0;
   
            if (changes > 0)
                VB_PRINTF(0, "Repeating analysis passes.\n");
   
        } while (changes > 0);
        
        VB_PRINTF(0, ("> %d words marked invalid during analysis\n", 
                  tot_changes));
    }

    /* -------------------------------------------------------------------- */
    /*  Generate labels as appropriate.                                     */
    /* -------------------------------------------------------------------- */
    VB_PRINTF(0, "Generating labels...\n");
    generate_labels();

    /* -------------------------------------------------------------------- */
    /*  Generate the disassembled text for each valid instruction.          */
    /* -------------------------------------------------------------------- */
    VB_PRINTF(0, "Generating disassembled text...\n");
    generate_text();
}

/* ======================================================================== */
/* ======================================================================== */
/*  INSTRUCTION PRINTER FUNCTIONS                                           */
/* ======================================================================== */
/* ======================================================================== */
static char prt_buf[256], prt_buf2[256];

char *prt_err(uint_32 addr)     /* Internal error printer */
{
    sprintf(prt_buf, "Internal decode error at %.4X", addr);
    return strdup(prt_buf);
}

char *prt_imp(uint_32 addr)     /* Implied operands     */
{
    return strdup(mnemonic[instr[addr].mnemonic]);
}

char *prt_jsr(uint_32 addr)     /* JSR reg, label       */
{
    sprintf(prt_buf, "%-7s R%d,     %s",
            mnemonic[instr[addr].mnemonic], 
            instr[addr].op1.op,
            symtab_getsym(symtab, instr[addr].op2.op, 'X', 0));

    return strdup(prt_buf);
}

char *prt_cbr(uint_32 addr)     /* BRANCH label         */
{
    sprintf(prt_buf, "%-7s %s",
            mnemonic[instr[addr].mnemonic], 
            symtab_getsym(symtab, instr[addr].op1.op, 'X', 0));

    return strdup(prt_buf);
}

char *prt_bxt(uint_32 addr)     /* BEXT cond, label     */
{
    sprintf(prt_buf, "%-7s $%.1X,     %s",
            mnemonic[instr[addr].mnemonic], 
            instr[addr].op1.op,
            symtab_getsym(symtab, instr[addr].op2.op, 'X', 0));

    return strdup(prt_buf);
}

char *prt_dir(uint_32 addr)     /* OP label, reg        */
{
    sprintf(prt_buf2, "%s,", 
            symtab_getsym(symtab, instr[addr].op1.op, 'X', 0));

    sprintf(prt_buf, "%-7s %-8sR%d",
            mnemonic[instr[addr].mnemonic], 
            prt_buf2, instr[addr].op2.op);

    return strdup(prt_buf);
}


char *prt_imm(uint_32 addr)     /* OP #imm, reg         */
{
    if (instr[addr].op1.flags & OPF_ADDR)
    {
        sprintf(prt_buf2, "#%s,", 
                symtab_getsym(symtab, instr[addr].op1.op, 'X', 0));
    } else
    {
        sprintf(prt_buf2, "#$%.4X,", instr[addr].op1.op);
    }

    sprintf(prt_buf, "%-7s %-8sR%d", 
            mnemonic[instr[addr].mnemonic], 
            prt_buf2, instr[addr].op2.op);

    return strdup(prt_buf);
}


char *prt_imo(uint_32 addr)     /* OP reg, #imm         */
{
    sprintf(prt_buf, "%-7s R%d,     %s", 
            mnemonic[instr[addr].mnemonic], 
            instr[addr].op1.op, instr[addr].op2.op);

    return strdup(prt_buf);
}

char *prt_2rg(uint_32 addr)     /* OP reg, reg          */
{
    sprintf(prt_buf, "%-7s R%d,     R%d", 
            mnemonic[instr[addr].mnemonic], 
            instr[addr].op1.op, instr[addr].op2.op);

    return strdup(prt_buf);
}

char *prt_1rg(uint_32 addr)     /* OP reg               */
{
    sprintf(prt_buf, "%-7s R%d", 
            mnemonic[instr[addr].mnemonic], instr[addr].op1.op);

    return strdup(prt_buf);
}

char *prt_dcl(uint_32 addr)     /* DECLE [up to four words] */
{
    if (instr[addr].len > 4)
    {
        fprintf(stderr, "ERROR: DECLE w/ len > 4 at address %.4X\n", addr);
        exit(1);
    }
    if (instr[addr].len <= 0)
    {
        fprintf(stderr, "ERROR: DECLE w/ len <= 0 at address %.4X\n", addr);
        exit(1);
    }

    switch (instr[addr].len)
    {
        case 1: 
            sprintf(prt_buf, "DECLE   $%.4X", GET_WORD(addr));
            break;
        case 2: 
            sprintf(prt_buf, "DECLE   $%.4X,  $%.4X",
                    GET_WORD(addr    ), GET_WORD(addr + 1));
            break;
        case 3: 
            sprintf(prt_buf, "DECLE   $%.4X,  $%.4X,  $%.4X",
                    GET_WORD(addr    ), GET_WORD(addr + 1), 
                    GET_WORD(addr + 2));
            break;
        case 4: 
            sprintf(prt_buf, "DECLE   $%.4X,  $%.4X,  $%.4X,  $%.4X",
                    GET_WORD(addr    ), GET_WORD(addr + 1), 
                    GET_WORD(addr + 2), GET_WORD(addr + 3));
            break;
    }

    return strdup(prt_buf);
}

char *prt_bid(uint_32 addr)     /* BIDECLE [up to two bidecles] */
{
    /* implementation note:  BIDECLE relies on op1, op2 having been set up  */
    /* rather than reading from the memory image.                           */
    switch (instr[addr].len)
    {
        case 2:
            sprintf(prt_buf, "BIDECLE $%.4X", instr[addr].op1.op);
            break;

        case 4:
            sprintf(prt_buf, "BIDECLE $%.4X,  $%.4X",
                    instr[addr].op1.op,
                    instr[addr].op2.op);
            break;

        default:
            fprintf(stderr, "ERROR: BIDECLE w/ len != 2 or 4 at addr %.4X\n", 
                    addr);
            exit(1);
    }

    return strdup(prt_buf);
}

/* ======================================================================== */
/* ======================================================================== */
/*  MAIN PROGRAM, GENERIC CRAPOLA(TM)                                       */
/* ======================================================================== */
/* ======================================================================== */


static const char rcs_id[] UNUSED = "$Id$";

static struct option long_opts[] =
{
    {   "allow-branch-target-wrap",             0,      NULL,       'b'     },
    {   "allow-suspicious-pc-math",             0,      NULL,       'p'     },
    {   "allow-rare-opcodes",                   0,      NULL,       'r'     },
    {   "allow-hlt",                            0,      NULL,       'H'     },
    {   "disable-analysis",                     0,      NULL,       'a'     },
    {   "verbose",                              0,      NULL,       'v'     },

    {   "help",                                 0,      NULL,       'h'     },
    {   "?",                                    0,      NULL,       '?'     },
    {   "license",                              0,      NULL,       'l'     },
                                               
    {   NULL,                                   0,      NULL,       0       }
};

static const char *optchars = "bprhdv:h?l";

extern char *optarg;
extern int  optind, opterr, optopt;


/* ======================================================================== */
/*  USAGE            -- Just give usage info and exit.                      */
/* ======================================================================== */
void usage(void)
{
    fprintf(stderr, 
                                                                          "\n"
    "ROM_MERGE"                                                           "\n"
    "Copyright 2003, Joseph Zbiciak"                                      "\n"
                                                                          "\n"
    "Usage: \n"
    "    rom_merge [flags] file0.rom [file1.rom [file2.rom [...]]] out.rom\n"
    "\n"
    "The output .ROM file may safely overlap one of the input ROMs if -f is\n"
    "specified.  Otherwise, the output file may not overlap any of the"   "\n"
    "input ROMs as a safety precaution."                                  "\n"
                                                                          "\n"
    "Flags:"                                                              "\n"
    "    -f  --force     Allow output file to overwrite an input file."   "\n"
    "    -r  --replace   Allow later .ROMs to replace portions mapped by" "\n"
    "                    previous .ROMs.  (Default is error.)"            "\n"
    "    -l  --license   License information"                             "\n"
    " -h -?  --help      This usage info"                                 "\n"
                                                                          "\n"
    );

    exit(0);
}

/* ======================================================================== */
/*  LICENSE          -- Just give license/authorship info and exit.         */
/* ======================================================================== */
void license(void)
{
    fprintf(stderr, 
                                                                          "\n"
    "ROM_MERGE"                                                           "\n"
    "Copyright 2003, Joseph Zbiciak"                                      "\n"
                                                                          "\n"
    "This program is free software; you can redistribute it and/or modify""\n"
    "it under the terms of the GNU General Public License as published by""\n"
    "the Free Software Foundation; either version 2 of the License, or"   "\n"
    "(at your option) any later version."                                 "\n"
                                                                          "\n"
    "This program is distributed in the hope that it will be useful,"     "\n"
    "but WITHOUT ANY WARRANTY; without even the implied warranty of"      "\n"
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU"   "\n"
    "General Public License for more details."                            "\n"
                                                                          "\n"
    "You should have received a copy of the GNU General Public License"   "\n"
    "along with this program; if not, write to the Free Software"         "\n"
    "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA."           "\n"
                                                                          "\n"
    "Run \"rom_merge --help\" for usage information."                     "\n"
                                                                          "\n"
    );

    exit(0);
}


/* ======================================================================== */
/*  MAIN             -- In The Beginning, there was MAIN, and C was with    */
/*                      CONST and VOID, and Darkness was on the face of     */
/*                      the Programmer.                                     */
/* ======================================================================== */
int main(int argc, char *argv[])
{
    int c, option_idx = 0, value;
    char *output_file;
    int i;

    /* -------------------------------------------------------------------- */
    /*  Parse command-line arguments.                                       */
    /* -------------------------------------------------------------------- */
    while ((c = getopt_long(argc, argv, optchars, long_opts, &option_idx)) 
            != EOF)
    {
        value = 1;
        if (optarg)
            value = atoi(optarg);

        switch (c)
        {
            case 'f': force_overwrite = 1;              break;
            case 'r': allow_replace   = 1;              break;
            case 'h': case '?': usage();                break;
            case 'l': license();                        break;
            default:
            {
                fprintf(stderr, "Unrecognized argument: '%c'\n", c);
                exit(1);
            }
        }
    }

    /* -------------------------------------------------------------------- */
    /*  Must have at least three additional arguments:  Two input files     */
    /*  and an output file.                                                 */
    /* -------------------------------------------------------------------- */
    if (optind + 3 > argc)
    {
        fprintf(stderr, "ERROR:  Must provide at least two input files and "
                        "an output file\n");
        exit(1);
    }

    /* -------------------------------------------------------------------- */
    /*  Initialize our main Intellicart image.  We work by merging all      */
    /*  others into this one.                                               */
    /* -------------------------------------------------------------------- */
    icartrom_init(&final_icart);

    /* -------------------------------------------------------------------- */
    /*  Remaining arguments are .ROM image filenames.  The first N-1 names  */
    /*  are input files.  The last name is the output file.  The output     */
    /*  file is not allowed to be the same as any of the input files names  */
    /*  unless the overwrite flag is used.                                  */
    /*                                                                      */
    /*  Note that I just do a simple string compare, and so this mechanism  */
    /*  is easy to fool.  Under UNIX, I could do safer checks (eg. stat()   */
    /*  and look at device and inode number), but that's overkill, and not  */
    /*  very portable.                                                      */
    /* -------------------------------------------------------------------- */
    output_file = argv[argc - 1];

    for (i = optind; i < argc - 1; i++)
    {
        /* ---------------------------------------------------------------- */
        /*  Check for accidental over write of output file.                 */
        /* ---------------------------------------------------------------- */
        if (!force_overwrite && !strcmp(argv[i], output_file))
        {
            fprintf(stderr, "ERROR:  Input file '%s' overwrites output file\n"
                            "        Use \"-f\" flag to force overwrite.\n",
                            argv[i]);
            exit(1);
        }

        /* ---------------------------------------------------------------- */
        /*  Re-initialize the temporary Intellicart.                        */
        /* ---------------------------------------------------------------- */
        icartrom_init(&temp_icart);

        /* ---------------------------------------------------------------- */
        /*  Read in the requested file.                                     */
        /* ---------------------------------------------------------------- */
        icartrom_readfile(argv[i], &temp_icart);

        /* ---------------------------------------------------------------- */
        /*  Merge this image into our final image.                          */
        /* ---------------------------------------------------------------- */
        merge_icarts(&final_icart, &temp_icart, allow_replace);
    }

    /* -------------------------------------------------------------------- */
    /*  Now, write the final file.                                          */
    /* -------------------------------------------------------------------- */
    icartrom_writefile(output_file, &final_icart);

    return 0;
}

/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 2002-2003, Joseph Zbiciak                  */
/* ======================================================================== */
